@@ -7,10 +7,6 @@ XAO::DO::Cache::Memory - memory storage back-end for XAO::Cache
 You should not use this object directly, it is a back-end for
 XAO::Cache.
 
- if($backend->exists(\@c)) {
-     return $backend->get(\@c);
- }
-
 =head1 DESCRIPTION
 
 Cache::Memory is the default implementation of XAO::Cache back-end. It
@@ -27,11 +23,12 @@ package XAO::DO::Cache::Memory;
 use strict;
 use XAO::Utils;
 use XAO::Objects;
+use Clone qw(clone);
 
 use base XAO::Objects->load(objname => 'Atom');
 
 use vars qw($VERSION);
-($VERSION)=(q$Id: Memory.pm,v 1.5 2004/02/23 21:44:43 am Exp $ =~ /(\d+\.\d+)/);
+$VERSION=(0+sprintf('%u.%03u',(q$Id: Memory.pm,v 2.1 2005/01/13 22:34:34 am Exp $ =~ /\s(\d+)\.(\d+)\s/))) || die "Bad VERSION";
 
 ###############################################################################
 
@@ -110,21 +107,18 @@ sub drop ($@) {
 
 ###############################################################################
 
-=item exists (@)
+=item drop_all ($)
 
-Checks if an element exists in the cache. Does not update its access
-time, but checks it. If the element should be expired it removes it from
-the cache and returns false.
+Drops all elements.
 
 =cut
 
-sub exists ($$) {
-    my $self=shift;
+sub drop_all ($$$) {
+    my ($self,$key,$ed)=@_;
 
-    my $key=$self->make_key($_[0]);
-    my $ed=$self->{data}->{$key};
-
-    return $ed && $ed->{access_time} + $self->{expire} > time;
+    $self->{'data'}={ };
+    $self->{'least_recent'}=$self->{'most_recent'}=undef;
+    $self->{'current_size'}=0;
 }
 
 ###############################################################################
@@ -140,10 +134,16 @@ sub get ($$) {
     my $self=shift;
 
     my $key=$self->make_key($_[0]);
-    my $ed=$self->{data}->{$key} ||
-        throw $self "get - no such element in the cache ($key), internal error";
 
-    return $ed->{element};
+    ### dprint "MEMORY: get(",$key,")";
+
+    my $ed=$self->{'data'}->{$key};
+
+    my $expire=$self->{'expire'};
+
+    my $exists=($ed && (!$expire || $ed->{'access_time'} + $expire > time));
+
+    return $exists ? $ed->{'element'} : undef;
 }
 
 ###############################################################################
@@ -171,22 +171,27 @@ depends on when an element was accessed last.
 
 sub put ($$$) {
     my $self=shift;
+
     my $key=$self->make_key(shift);
-    my $element=shift;
+
+    # We store a deep copy, not an actual data piece. It must be OK to
+    # modify the original data after it's cached.
+    #
+    my $element=clone(shift);
 
     my $data=$self->{data};
     my $size=$self->{size};
     my $nsz=$size ? $self->calculate_size($element) : 0;
 
     my $lr=$self->{least_recent};
-    my $expire=$self->{expire};
+    my $expire=$self->{'expire'};
     my $now=time;
     my $count=5;
     while(defined($lr)) {
         my $lred=$data->{$lr};
         last unless $count--;
         last unless ($size && $self->{current_size}+$nsz>$size) ||
-                    $lred->{access_time}+$expire < $now;
+                    ($expire && $lred->{access_time}+$expire < $now);
         $lr=$self->drop_oldest($lr,$lred);
     }
 
@@ -198,6 +203,8 @@ sub put ($$$) {
         next        => $self->{most_recent},
     };
 
+    ### dprint "MEMORY: put(",$key," => ",$element,") size=",$self->{'size'}," expire=",$self->{'expire'};
+
     $data->{$self->{most_recent}}->{previous}=$key
         if defined($self->{most_recent});
 
@@ -205,7 +212,7 @@ sub put ($$$) {
     $self->{least_recent}=$key unless defined($self->{least_recent});
     $self->{current_size}+=$nsz;
 
-    undef;
+    return undef;
 }
 
 ###############################################################################
@@ -220,10 +227,10 @@ sub setup ($%) {
     my $self=shift;
     my $args=get_args(\@_);
 
-    $self->{data}={};
-    $self->{current_size}=0;
-    $self->{expire}=$args->{expire} || 60;
-    $self->{size}=$args->{size} || 0;
+    $self->{'expire'}=$args->{'expire'} || 0;
+    $self->{'size'}=($args->{'size'} || 0) * 1024;
+
+    $self->drop_all();
 }
 
 ###############################################################################
@@ -245,6 +252,8 @@ Drops oldest element from the cache using supplied key and element.
 sub drop_oldest ($$$) {
     my ($self,$key,$ed)=@_;
 
+    ### dprint "drop_oldest()";
+
     $self->{most_recent}=undef if defined($self->{most_recent}) &&
                                   $self->{most_recent} eq $key;
 
@@ -259,7 +268,7 @@ sub drop_oldest ($$$) {
 
     delete $data->{$key};
 
-    # $self->print_chain();
+    ### $self->print_chain();
 
     return $previous;
 }
@@ -328,7 +337,7 @@ sub touch ($$$) {
         $self->{most_recent}=$data->{$ed->{next}}->{previous}=$key;
     }
 
-    # $self->print_chain;
+    ### $self->print_chain;
 }
 
 ###############################################################################
